open Async.Std
open Replica
type 'a server_status = Returned of 'a | Crashed
(* server_id -> max delay time of the server (in second) *)

let max_delay : (int, float) Hashtbl.t = Hashtbl.create 17
let crash_info : (int, int) Hashtbl.t = Hashtbl.create 17
(* wait for a random time, up to the max delay of the server *)
let random_delay ser_id =
  let mdelay = Hashtbl.find max_delay ser_id in
  if mdelay > 0. then
    after (Core.Std.sec (Random.float (Random.float mdelay)))
  else return ()

(* a protocol using Async.Std.Pipe. Simulates network condition *)
module P : sig
  include Protocol
  val create : unit -> reader * writer
end = struct
  type event = Vote of int * value
             | Collect of int
             | Decide of value
             | Connect
  type msg   = int * event
  type reader = msg Pipe.Reader.t
  type writer = msg Pipe.Writer.t
  let create () : reader * writer = Pipe.create ()

  (* error handling read *)
  let receive r  = try Pipe.read r
                  with _ ->
                  printf "Exception Received While Reading\n%!"; return `Eof
  (* error handling write *)
  let send w (sid, e) =
    let crash = Hashtbl.find crash_info sid - 1 in
    match e with
    |Collect r | Vote (r, _) when crash >= 0 && r >= crash ->
      ()
    | _ ->
    random_delay sid >>> fun () ->
    try Pipe.write_without_pushback w (sid,e) with
    | _ -> printf "Exception Received While Sending\n%!"

  let to_string msg =
    let sid, f = fst msg, string_of_val in
    match snd msg with
    | Vote (i,v) -> Printf.sprintf "Server %d: Vote(%d,%s)" sid i (f v)
    | Collect i  -> Printf.sprintf "Server %d: Collect(%d)" sid i
    | Decide v   -> Printf.sprintf "Server %d: Decide(%s)"  sid (f v)
    | Connect    -> Printf.sprintf "Server %d: Connect"     sid
end

(***********************local control server setup ****************************)
(* machine id *)
type id = int
(* the channel of communication between replicas *)
type pipe = P.reader * P.writer
(* each server has an id and a list of communications *)
type server = {id : int; socket : pipe list}
(* configuration of each server. [delay] is the max delay time of communication
 * [crash] > 0 means that the server will be shut down immediately as the
 * protocol starts, and thus is considered crashed *)
type conf = { ini: int; delay: int; crash:int}

(* set up k servers *)
let server_up (k : int) : server list =
  (* set up a single server, note that channels with servers
     before this id are already created and just passed along *)
  let rec create_pipes count id : pipe option list =
    let get i : pipe option = (
      if i = id then None else Some (P.create ())
    ) in
    if count >= k then []
    else (get count) :: create_pipes (count + 1) id
  in
  (* matrix !! *)
  (* helper to create the connections for all k servers. Each replica does not
   * connect with itself *)
  let rec create_pipe_matrix cur_id =
    if cur_id >= k then []
    else (create_pipes 0 cur_id)::(create_pipe_matrix (cur_id + 1))
  in
  let matrix = create_pipe_matrix 0 in
  (* Give the writer and reader side of pipe to respective server *)
  let construct_server id =
    let extract_w x = match x with
      |None -> failwith "cannot extract writer"
      |Some (_, w) -> w
    in
    let extract_r x = match x with
      |None -> failwith "cannot extract reader"
      |Some (r, _) -> r
    in
    let all_reader_list = List.nth matrix id in
    let reader_list = List.filter (fun x -> x <> None) all_reader_list in
    let readers = List.map extract_r reader_list in
    let map_fun i l = if i = id then None else List.nth l id in
    let all_writer_list = List.mapi map_fun matrix in
    let writer_list = List.filter (fun x -> x <> None) all_writer_list in
    let writers = List.map extract_w writer_list in
    let pairs = List.combine readers writers in
    {id = id; socket = pairs}
  in
  let rec all_up i =
    if i >= k then []
    else (construct_server i) :: (all_up (i+1))
  in
  all_up 0

(* read the configuration file and obtain info about delay and crash *)
let config config_file =
  Reader.file_lines config_file >>= fun conflst ->
  match conflst with
  | [] -> failwith "Not enough number of lines in config file."
  | h::t ->
    let f = int_of_string h in
    if (List.length t) < 3*f + 1 then failwith "Not enough server configurations."
    else
      List.map (Str.split (Str.regexp "[ \t]+")) t
      |> List.map (List.map (fun str -> int_of_string (String.trim str)))
      |> List.mapi (fun i -> function
          | [ini ; delay; crash] ->
            Hashtbl.replace max_delay i (float_of_int delay);
            Hashtbl.replace crash_info i crash;
            (ini, crash)
          | _ -> printf "Invalid configuration format at line %d" i; (0,0))
      |> fun l -> return (f, l)

let main config_file () =
  (* read from the config file *)
  config config_file >>=
  fun (f, inits) ->
  let servers = server_up (3*f + 1) in
  (* given a server [s], config the server and start running [s] *)
  let run_replica s =
      let module Config = struct
        let server_id = s.id
        let total = 3*f + 1
        let f     = f
        let init_val = fst (List.nth inits s.id)
                      |> string_of_int |> val_of_string
        let logger = printf "[Server %d] %s" s.id
      end in
      let module Rep = Make (P) (Config) in
      let user_return = Ivar.create () in
      let server_return = Ivar.create () in
      (* write decision to consensus.out *)
      let append text =
          let outfile = "consensus.out" in
          Writer.with_file outfile ~append:true ~f:(fun writer ->
          Writer.write writer text;
          return ())
      in
      let output_handler=
        Ivar.read server_return >>= fun v ->
        (* Write message to file *)
        let string_to_write =
        (match v with
        |Crashed -> "server crashed"
        |Returned x -> string_of_val x) in
        append
        (Printf.sprintf "Machine[%d] decides on %s\n" s.id string_to_write)
      in
      (* if user decides, fill in the value on the controller side *)
      let _ = Ivar.read user_return >>= fun v ->
              return (Ivar.fill_if_empty server_return (Returned v)) in
      don't_wait_for (
        let _ =
          if snd (List.nth inits s.id) > 0
          then Ivar.fill_if_empty server_return Crashed
          else ()
        in Rep.run s.socket user_return
      );
      output_handler >>= fun _ ->
      return ()
  in
  (* if all replicas have either decided or crashed, exit the program *)
  Deferred.List.all (List.map run_replica servers) >>= fun _ -> exit 0

let () = Command.async_basic
    ~summary:"."
    Command.Spec.(
      empty
      +> anon ("config_file" %:  string)
    ) main |> Command.run
