open Async.Std
open Replica

(* Global states for injected random delays *)
let has_random_delay = ref false

(* preset crash time for each server, -1 means it doesnt crash *)
let crash_table = Hashtbl.create 17

(* has_crashed[idx] = true iff server with id [idx] is crashed. *)
let has_crashed = Hashtbl.create 17

(* Maximum time delayed *)
let max_delay = ref 0.2

let random_delay () =
  if !has_random_delay then
    let delay = Random.float !max_delay in
    after (Core.Std.sec (Random.float delay))
  else return ()

(* Protocol and logger methods. *)
module P : Protocol with type reader = Reader.t
                     and type writer = Writer.t = struct
  type event = Vote of int * value
             | Collect of int
             | Decide of value
             | Connect

  type msg    = int * event
  type reader = Reader.t
  type writer = Writer.t

  (* [to_string msg] will return a string representation of the message [msg] *)
  let to_string (sid, e) =
    match e with
    | Vote (i,v) -> Printf.sprintf "(%d, Vote(%d,%s))" sid i (string_of_val v)
    | Collect i  -> Printf.sprintf "(%d, Collect(%d))" sid i
    | Decide v   -> Printf.sprintf "(%d, Decide(%s))"  sid (string_of_val v)
    | Connect    -> Printf.sprintf "(%d, Connect)"     sid

  (* [receive r] will return a defferred that either carries a message from
     reader [r] or return `Eof if error encountered. The message will be
     determined after configurated random delays. *)
  let receive r =
    if not (Reader.is_closed r) then
      random_delay () >>= fun _ ->
      try (Reader.read_marshal r >>= fun msg ->
           (match msg with
            | `Eof  -> printf "Received end of server.\n%!"
            | `Ok v -> printf "Received %s\n%!" (to_string v)
           ); return msg)
      with | _ -> print_endline "Exception received while reading."; return `Eof
    else return `Eof

  (* [send w msg] will send message [msg] using writer [w].
     The message will be sent after the configurated random delays.
     Will block all the messages if the replica reaches the specific round
     indicated by the configuraiton file. *)
  let send w (id, e) =
    if Writer.is_closed w then () else
    let crash_round = Hashtbl.find crash_table id in
    match e with
    | Collect r | Vote (r,_) when r = crash_round - 1 ->
      Hashtbl.replace has_crashed id true;
      printf "Machine %d crashed at round %d\n%!" id r
    | Collect r | Vote (r,_) when crash_round > 0 && r > crash_round -> ()
    | _ ->
      if try Hashtbl.find has_crashed id with _ -> false then
        printf "Server %d crashed. No more message out.\n%!" id
      else
        random_delay () >>> fun _ ->
        printf "Send: %s\n%!" (to_string (id, e));
        try (Writer.write_marshal w [] (id, e)) with
        _ -> printf "Unable to send message. \n%!"
end

(* [get_connections sid] will connects to all the servers specified in
   file "addresses.txt", regarding the ([sid] - 1)th address as the address
   of current replica. Return a list of [3*f] reader writer list *)
let get_connections idx : (Reader.t * Writer.t) list Deferred.t =
  let split_host_and_port s =
    match Str.split (Str.regexp_string ":") s with
    | [host; port] -> Some (host, int_of_string port)
    | _            -> (printf "Error: invalid host:port string %s%!" s; None)
  in

  let to_host_and_port s =
    match split_host_and_port s with
    | None -> None
    | Some (host, port) -> Some (Tcp.to_host_and_port host port)
  in

  (* [try_connect s] takes a address and port, returns a connected TCP pipe. *)
  let rec try_connect i s : (Reader.t * Writer.t) option Deferred.t =
    match to_host_and_port s with
    | None -> failwith "Invalid address string"
    | Some hostport -> begin
        try_with (fun () -> Tcp.connect hostport) >>= function
        | Core.Std.Ok (_, r, w) ->
          (printf "TCP connect to server %d\n%!" i); 
          return (Some (r, w))
        | Core.Std.Error _ -> 
          (printf "Failed to connect to %s\n%!" s);
          return None
      end
  in

  (* [set_up_tcp port tbl] will set up a TCP servers. This listener will be
     invoked once when the first connection comes, and save the reader writer
     into hashtble [tbl]. The listener will never returns so that we could hold
     onto the reader and writer. *)
  let set_up_tcp (port:int) (table : (int, 'a Ivar.t) Hashtbl.t) : unit Deferred.t =
    let setup _ r w =
      let open P in
      receive r >>= function
      | `Eof -> printf "Should have the connection event.\n%!"; return ()
      | `Ok (i, Connect) ->
        printf "TCP connection received from %d\n%!" i;
        Ivar.fill_if_empty (Hashtbl.find table i) (r,w); never ()
      | `Ok (i, _) ->
        printf "TCP connection received other events from server %d. \n %!" i;
        Ivar.fill_if_empty (Hashtbl.find table i) (r,w); never ()
    in
    ignore (Tcp.Server.create (Tcp.on_port port) setup);
    printf "Successfully set up the TCP server %d on port %d\n%!" idx port;
    never ()
  in

  Reader.file_lines "addresses.txt" >>= fun addresses ->

  (* Create a hashtable for each server in the group except the current ont. *)
  let wrtbl = Hashtbl.create 17 in
  List.iteri (fun i _ ->
    if i <> idx then Hashtbl.add wrtbl i (Ivar.create ()) else ()
  ) addresses;

  (* Find the port in the machine. *)
  let port =
    List.nth addresses idx
    |> split_host_and_port
    |> function
    | None -> failwith "Invalid address."
    | Some (_, p) -> p
  in

  (* Try to connect to any other servers *)
  let ret_lst = List.mapi (fun i s ->
      if i <> idx then
        try_connect i s >>= (function
            | Some (r,w) -> 
              P.send w (idx, P.Connect);
              Ivar.fill_if_empty (Hashtbl.find wrtbl i) (r,w);
              return ()
            | None -> return ())
      else return ()
    ) addresses in
  Deferred.all ret_lst >>= fun _ ->

  (* Set up the TCP listener to receive TCP connections *)
  ignore(set_up_tcp port wrtbl);

  (* Return a list of Reader Writer Deferred.t *)
  List.mapi (fun i _ -> i) addresses
  |> List.filter (fun i -> i<>idx)
  |> List.map (fun i -> Ivar.read (Hashtbl.find wrtbl i))
  |> Deferred.List.all

(* Start the [idx+1]th server with first vote [init]
   [rand] is maximum seconds the message of this replica could be dealyed 
   [crash] if [crash] <> 0, then this replica will be blocked at round [crash-1]*) 
let main idx init rand crash () =
  Hashtbl.replace crash_table idx crash;
  get_connections idx >>= fun connections ->
  if rand = 0 then () else
    has_random_delay := true; max_delay := (float_of_int rand);
  let l = List.length connections in
  if l mod 3 <> 0 then failwith "Invalid number of servers in the group."
  else printf "Enter the main protocol.\n%!";
  let module Conf = struct
    let server_id = idx
    let f         = (l - l mod 3) / 3
    let total     = l + 1
    let init_val  = Replica.val_of_string init
    let logger    = printf "%s\n%!" 
  end in
  let module Rep = Replica.Make (P) (Conf) in
  let ret = Ivar.create () in
  let output_handler =
    Ivar.read ret >>= fun v ->
    let append text =
      Writer.with_file "consensus.out" ~append:true
        ~f:(fun writer -> Writer.write writer text; return ())
    in
    append (Printf.sprintf "Machine[%d] decides on %s\n" idx (string_of_val v))
  in
  Deferred.both (Rep.run connections ret) output_handler >>= fun _ ->
  exit 0

let () = Command.async_basic
  ~summary:"Run the [i]th replica."
  Command.Spec.(
    empty
    +> anon ("Server id." %: int)
    +> anon ("Initial value passed into round 0." %:  string)
    +> anon ("Random delay time" %: int)
    +> anon ("Crash round. 0 if not crashed." %: int)
  ) main |> Command.run
