open Async.Std

(* Serializable value to decide on. *)
type value
val string_of_val : value -> string
val val_of_string : string -> value

(* Protocol specifying the way messages are sent across networks. *)
module type Protocol = sig
  type event = Vote of int * value  (* Vote    (round number, value to be sent) *)
             | Collect of int       (* Collect (round number) *)
             | Decide of value      (* Decide  (value decided) *)
             | Connect

  (* Represent the message sending across the internet: (server id, event) *)
  type msg   = int * event
  type reader
  type writer

  (* [send wr msg] will send a message [msg] using writer [wr] *)
  val send      : writer -> msg -> unit

  (* [receive re] will return a deferred that will be filled when the next
     message from reader [re] was received. *)
  val receive   : reader -> [`Ok of msg |`Eof] Deferred.t

  (* [to_string msg] will serialize the [msg] to string. *)
  val to_string : msg -> string
end

(* Module that configurate the replica. *)
module type Config = sig
  val  server_id : int             (* Server ID *)
  val  total     : int             (* Total number of servers *)
  val  f         : int             (* Tolerance *)
  val  init_val  : value           (* Initial value of this replica *)
  val  logger    : string -> unit  (* Make a log to standard output. *)
end

(* Standard module type for a replica. *)
module type S = sig
  type reader
  type writer

  (* [run plst ret] will bring up a replica, that connects to other replicas via
  the list of reader and writers [lst]. This replica will conduct the
  2/3 consensus protocol. When it decides, it fills in the Ivar [ret]. Run does
  not necessarily return after the replica decides, because it is important
  that it stays in the network of replicas when others ask for its opinion *)
  val run : (reader * writer) list -> value Ivar.t -> unit Deferred.t
end

(* Module to create a replica *)
module Make (P : Protocol) (C : Config)
     : S with type reader  = P.reader
          and type writer  = P.writer
